@@ -757,6 +757,14 @@ TEST_CASE("Logic program", "[asp]") {
 		REQUIRE(ctx.varInfo(lp.getLiteral(g).var()).frozen());
 	}
 
+	SECTION("testExternalsAreFrozen") {
+		lpAdd(lp.start(ctx),
+			"#external a.\n");
+		REQUIRE(lp.endProgram());
+		REQUIRE(lp.getLiteral(a).var() == 1);
+		REQUIRE(ctx.varInfo(lp.getLiteral(a).var()).frozen());
+	}
+
 	SECTION("testComputeTrueBug") {
 		lpAdd(lp.start(ctx),
 			"a :- not b.\n"
@@ -1148,6 +1156,58 @@ TEST_CASE("Logic program", "[asp]") {
 		REQUIRE(lp.getLiteral(b) == lit_false());
 		REQUIRE(t.numAtoms() == 0);
 	}
+
+	SECTION("testTheoryHeadEvenIfRuleIsDropped") {
+		lpAdd(lp.start(ctx),
+			":- b.\n"
+			"a :- b, c.\n");
+		Potassco::TheoryData& t = lp.theoryData();
+		t.addTerm(0, "ta");
+		t.addTerm(1, "tc");
+		t.addAtom(a, 0, Potassco::toSpan<Potassco::Id_t>());
+		t.addAtom(c, 1, Potassco::toSpan<Potassco::Id_t>());
+		REQUIRE(t.numAtoms() == 2);
+		lp.endProgram();
+		REQUIRE(lp.getLiteral(a) == lit_false());
+		REQUIRE(lp.getLiteral(b) == lit_false());
+		REQUIRE(lp.getLiteral(c) != lit_false());
+		REQUIRE(t.numAtoms() == 1);
+	}
+
+	SECTION("testTheoryHeadEvenIfRuleIsDroppedLater") {
+		lpAdd(lp.start(ctx),
+			"{c}.\n"
+			"b :- c.\n"
+			":- c.\n");
+		Potassco::TheoryData& t = lp.theoryData();
+		t.addTerm(0, "x");
+		t.addTerm(1, "c");
+		Potassco::Id_t term = 1;
+		t.addElement(0, Potassco::toSpan(&term, 1), 0);
+		term = 0;
+		t.addAtom(b, 0, Potassco::toSpan(&term, 1));
+		REQUIRE(t.numAtoms() == 1);
+		lp.endProgram();
+		REQUIRE(lp.getLiteral(a) == lit_false());
+		REQUIRE(lp.getLiteral(b) == lit_false());
+		REQUIRE(lp.getLiteral(c) == lit_false());
+		REQUIRE(t.numAtoms() == 0);
+	}
+
+	SECTION("testTheoryHeadEvenIfRuleIsSkipped") {
+		lpAdd(lp.start(ctx),
+			"{a}.\n"
+			"b :- a, not a.\n");
+		Potassco::TheoryData& t = lp.theoryData();
+		t.addTerm(0, "ta");
+		t.addAtom(b, 0, Potassco::toSpan<Potassco::Id_t>());
+		REQUIRE(t.numAtoms() == 1);
+		lp.endProgram();
+		REQUIRE(lp.getLiteral(a) != lit_false());
+		REQUIRE(lp.getLiteral(b) == lit_false());
+		REQUIRE(t.numAtoms() == 0);
+	}
+
 	SECTION("testFalseBodyTheoryAtomsAreKept") {
 		lp.start(ctx);
 		a = lp.newAtom();
@@ -1504,6 +1564,37 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		REQUIRE(findSmodels(exp, lp));
 	}
 
+	SECTION("testSimplifyRuleEq") {
+		lp.start(ctx);
+		lp.updateProgram();
+		lpAdd(lp,
+			"{x3}.\n"
+			"x1 :- x3.\n"
+			"x2 :- x3.\n");
+
+		// x1 == x2
+		REQUIRE((lp.endProgram() && ctx.endInit()));
+		uint32 x1 = lp.getAtom(1)->id();
+		uint32 x2 = lp.getAtom(2)->id();
+		REQUIRE(x1 == x2);
+
+		lp.updateProgram();
+
+		lpAdd(lp,
+			"{x4}.\n"
+			"x5 :- 2{x1, x2, x4}.");
+		REQUIRE(lp.endProgram());
+		REQUIRE(lp.getAtom(5)->supports() == 1);
+		PrgBody* body = lp.getBody(lp.getAtom(5)->supps_begin()->node());
+		REQUIRE(body->type() == Potassco::Body_t::Sum);
+		REQUIRE(body->bound() == 2);
+		REQUIRE(body->size() == 2);
+
+		uint32 eqIdx = body->goal(0).var() == 4;
+		REQUIRE(body->weight(eqIdx) == 2);
+		REQUIRE(body->weight(1 - eqIdx) == 1);
+	}
+
 	SECTION("testEqUnfreeze") {
 		lp.start(ctx);
 		lp.updateProgram();
@@ -1956,6 +2047,32 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		}
 		REQUIRE((!foundA && foundB));
 	}
+
+	SECTION("testWriteExternalBug") {
+		lp.start(ctx);
+		lp.updateProgram();
+		lpAdd(lp,
+			"#external a."
+			"#external b."
+			"#external c."
+			"#external d."
+			"b.");
+		lp.endProgram();
+		std::stringstream str;
+		AspParser::write(lp, str, AspParser::format_aspif);
+		int foundA = 0, foundB = 0, foundC = 0, foundD = 0;
+		for (std::string x; std::getline(str, x);) {
+			if (x.find("5 1 2") == 0) { ++foundA; }
+			if (x.find("5 2 2") == 0) { ++foundB; }
+			if (x.find("5 3 2") == 0) { ++foundC; }
+			if (x.find("5 4 2") == 0) { ++foundD; }
+		}
+		REQUIRE(foundA == 1);
+		REQUIRE(foundB == 0);
+		REQUIRE(foundC == 1);
+		REQUIRE(foundD == 1);
+	}
+
 	SECTION("testWriteUnfreeze") {
 		lp.start(ctx);
 		lp.updateProgram();
@@ -2046,6 +2163,7 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		lp.endProgram();
 		REQUIRE_FALSE(lp.isExternal(a));
 	}
+
 	SECTION("testAssumptionsAreVolatile") {
 		lp.start(ctx);
 		lp.updateProgram();
@@ -2061,6 +2179,14 @@ TEST_CASE("Incremental logic program", "[asp]") {
 		lp.getAssumptions(assume);
 		REQUIRE((assume.size() == 1 && assume[0] == ~lp.getLiteral(a)));
 	}
+
+	SECTION("testAssumptionsAreFrozen") {
+		lpAdd(lp.start(ctx), "{a}. #assume{a}.");
+		REQUIRE(lp.endProgram());
+		REQUIRE(lp.getLiteral(a).var() == 1);
+		REQUIRE(ctx.varInfo(lp.getLiteral(a).var()).frozen());
+	}
+
 	SECTION("testProjectionIsExplicitAndCumulative") {
 		lp.start(ctx);
 		lp.updateProgram();
