@@ -180,7 +180,7 @@ bool propagate(clingo_propagate_control_t *control, const clingo_literal_t *chan
   return true;
 }
 
-bool undo(clingo_propagate_control_t *control, const clingo_literal_t *changes, size_t size, propagator_t *data) {
+void undo(clingo_propagate_control_t *control, const clingo_literal_t *changes, size_t size, propagator_t *data) {
   // get the thread specific state
   state_t state = data->states[clingo_propagate_control_thread_id(control)];
 
@@ -194,8 +194,6 @@ bool undo(clingo_propagate_control_t *control, const clingo_literal_t *changes,
       state.holes[hole] = 0;
     }
   }
-
-  return true;
 }
 
 bool print_model(clingo_model_t const *model) {
@@ -304,6 +302,7 @@ int main(int argc, char const **argv) {
     (clingo_propagator_propagate_callback_t)propagate,
     (clingo_propagator_undo_callback_t)undo,
     NULL,
+    NULL,
   };
   // user data for the propagator
   propagator_t prop_data = { NULL, 0, NULL, 0 };
