@@ -36,26 +36,21 @@ namespace Gringo { namespace Input {
 
 // {{{ definition of Block
 
-Block::Block(Location const &loc, String name, IdVec &&params)
-    : loc(loc)
-    , name(name)
-    , params(std::move(params))
-    , edb(std::make_shared<Ground::SEdb::element_type>(nullptr, SymVec())) {
+Ground::SEdb Block::make_sig() const {
+    auto edb = std::make_shared<Ground::SEdb::element_type>(nullptr, SymVec());
     UTermVec args;
-    for (auto &param : this->params) { args.emplace_back(make_locatable<ValTerm>(param.first, Symbol::createId(param.second))); }
-    if (args.empty()) { std::get<0>(*edb) = make_locatable<ValTerm>(loc, Symbol::createId(name)); }
-    else              { std::get<0>(*edb) = make_locatable<FunctionTerm>(loc, name, std::move(args)); }
+    for (auto const &param : params) {
+        args.emplace_back(make_locatable<ValTerm>(param.first, Symbol::createId(param.second)));
+    }
+    if (args.empty()) {
+        std::get<0>(*edb) = make_locatable<ValTerm>(loc, Symbol::createId(name));
+    }
+    else {
+        std::get<0>(*edb) = make_locatable<FunctionTerm>(loc, name, std::move(args));
+    }
+    return edb;
 }
 
-Block::Block(Block&&) = default;
-Block &Block::operator=(Block&&) = default;
-Block::~Block() = default;
-
-Term const &Block::sig() const {
-    return *std::get<0>(*edb);
-}
-Block::operator Term const &() const { return sig(); }
-
 // }}}
 // {{{ definition of Program
 
@@ -63,10 +58,14 @@ Program::Program() {
     begin(Location("<internal>", 1, 1, "<internal>", 1, 1), "base", IdVec({}));
 }
 
-Program::Program(Program &&) = default;
+bool Program::empty() const {
+    return blocks_.empty() && sigs_.empty() && theoryDefs_.empty();
+}
 
 void Program::begin(Location const &loc, String name, IdVec &&params) {
-    current_ = &*blocks_.push(loc, (std::string("#inc_") + name.c_str()).c_str(), std::move(params)).first;
+    auto block = Block{loc, (std::string("#inc_") + name.c_str()).c_str(), std::move(params)};
+    auto sig = block.make_sig();
+    current_ = &blocks_.try_emplace(std::move(sig), std::move(block)).first.value();
 }
 
 void Program::add(UStm &&stm) {
@@ -78,13 +77,13 @@ void Program::add(UStm &&stm) {
 }
 
 void Program::addInput(Sig sig) {
-    sigs_.push(sig);
+    sigs_.insert(sig);
 }
 
 void Program::add(TheoryDef &&def, Logger &log) {
     auto it = theoryDefs_.find(def.name());
     if (it == theoryDefs_.end()) {
-        theoryDefs_.push(std::move(def));
+        theoryDefs_.insert(std::move(def));
     }
     else {
         GRINGO_REPORT(log, Warnings::RuntimeError)
@@ -96,17 +95,19 @@ void Program::add(TheoryDef &&def, Logger &log) {
 }
 
 void Program::rewrite(Defines &defs, Logger &log) {
-    for (auto &block : blocks_) {
-        // {{{3 replacing definitions
+    for (auto it = blocks_.begin(), ie = blocks_.end(); it != ie; ++it) {
+        auto &block  = it.value();
+        auto const &sig = it.key();
+        // replacing definitions
         Defines incDefs;
 
         UTermVec args;
         AuxGen gen;
-        for (auto &param : block.params) {
+        for (auto const &param : block.params) {
             args.emplace_back(gen.uniqueVar(param.first, 0, "#Inc"));
             incDefs.add(param.first, param.second, get_clone(args.back()), false, log);
         }
-        sigs_.push(Sig(block.name, static_cast<uint32_t>(args.size()), false));
+        sigs_.insert(Sig(block.name, static_cast<uint32_t>(args.size()), false));
         UTerm blockTerm(args.empty()
             ? (UTerm)make_locatable<ValTerm>(block.loc, Symbol::createId(block.name))
             : make_locatable<FunctionTerm>(block.loc, block.name, get_clone(args)));
@@ -114,7 +115,9 @@ void Program::rewrite(Defines &defs, Logger &log) {
         blockTerm->collect(blockBound, true);
         incDefs.init(log);
 
-        for (auto &fact : block.addedEdb) { sigs_.push(fact.sig()); }
+        for (auto &fact : block.addedEdb) {
+            sigs_.insert(fact.sig());
+        }
         auto replace = [&](Defines &defs, Symbol fact) -> Symbol {
             if (defs.empty() || fact.type() == SymbolType::Special) { return fact; }
             UTerm rt;
@@ -123,71 +126,96 @@ void Program::rewrite(Defines &defs, Logger &log) {
             if (rt) {
                 Location loc{rt->loc()};
                 block.addedStms.emplace_back(make_locatable<Statement>(loc, gringo_make_unique<SimpleHeadLiteral>(make_locatable<PredicateLiteral>(loc, NAF::POS, std::move(rt))), UBodyAggrVec{}));
-                return Symbol();
+                return {};
             }
-            else if (rv.type() != SymbolType::Special) { return rv; }
-            else { return fact; }
+            if (rv.type() != SymbolType::Special) {
+                return rv;
+            }
+            return fact;
         };
         if (!defs.empty() || !incDefs.empty()) {
             for (auto &fact : block.addedEdb) {
                 Symbol rv = replace(incDefs, replace(defs, fact));
-                if (rv.type() != SymbolType::Special) { std::get<1>(*block.edb).emplace_back(rv); }
+                if (rv.type() != SymbolType::Special) {
+                    std::get<1>(*sig).emplace_back(rv);
+                }
             }
             block.addedEdb.clear();
         }
-        else if (std::get<1>(*block.edb).empty()) { std::swap(std::get<1>(*block.edb), block.addedEdb); }
-        else { std::copy(block.addedEdb.begin(), block.addedEdb.end(), std::back_inserter(std::get<1>(*block.edb))); }
-        // {{{3 rewriting
+        else if (std::get<1>(*sig).empty()) {
+            std::swap(std::get<1>(*sig), block.addedEdb);
+        }
+        else {
+            std::copy(block.addedEdb.begin(), block.addedEdb.end(), std::back_inserter(std::get<1>(*sig)));
+        }
+        // rewriting
+        // steps:
+        // 1. unpool
+        // 2. initialize theory
+        // 3. simplify
+        // 4. shift
+        // 5. unpool comparisions
+        // 6. rewrite aggregates
         auto rewrite2 = [&](UStm &x) -> void {
-            std::get<1>(*block.edb).emplace_back(x->isEDB());
-            if (std::get<1>(*block.edb).back().type() == SymbolType::Special) {
+            std::get<1>(*sig).emplace_back(x->isEDB());
+            if (std::get<1>(*sig).back().type() == SymbolType::Special) {
                 x->add(make_locatable<PredicateLiteral>(block.loc, NAF::POS, get_clone(blockTerm), true));
                 x->rewrite();
                 block.stms.emplace_back(std::move(x));
-                std::get<1>(*block.edb).pop_back();
+                std::get<1>(*sig).pop_back();
+            }
+            else {
+                sigs_.insert(std::get<1>(*sig).back().sig());
             }
-            else { sigs_.push(std::get<1>(*block.edb).back().sig()); }
         };
         auto rewrite1 = [&](UStm &x) -> void {
             x->initTheory(theoryDefs_, log);
             if (x->simplify(project_, log)) {
-                if (x->hasPool(false)) { for (auto &y : x->unpool(false)) { rewrite2(y); } }
-                else                   { rewrite2(x); }
+                for (auto &y : x->unpoolComparison()) {
+                    rewrite2(y);
+                }
             }
         };
         for (auto &x : block.addedStms) {
             x->replace(defs);
             x->replace(incDefs);
             x->assignLevels(blockBound);
-            if (x->hasPool(true)) { for (auto &y : x->unpool(true)) { rewrite1(y); } }
-            else                  { rewrite1(x); }
+            if (x->hasPool()) {
+                for (auto &y : x->unpool()) {
+                    rewrite1(y);
+                }
+            }
+            else {
+                rewrite1(x);
+            }
         }
         block.addedStms.clear();
-        // }}}3
     }
-    // {{{3 projection
-    for (auto &x : project_) {
-        if (!x.done) {
-            Location loc(x.project->loc());
+    // projection
+    for (auto it = project_.begin(), ie = project_.end(); it != ie; ++it) {
+        if (!it.value().second) {
+            Location loc(it.value().first->loc());
             UBodyAggrVec body;
-            body.emplace_back(gringo_make_unique<SimpleBodyLiteral>(make_locatable<ProjectionLiteral>(loc, get_clone(x.project))));
+            body.emplace_back(gringo_make_unique<SimpleBodyLiteral>(make_locatable<ProjectionLiteral>(loc, get_clone(it.value().first))));
             stms_.emplace_back(make_locatable<Statement>(
                 loc,
-                gringo_make_unique<SimpleHeadLiteral>(make_locatable<PredicateLiteral>(loc, NAF::POS, get_clone(x.projected))),
+                gringo_make_unique<SimpleHeadLiteral>(make_locatable<PredicateLiteral>(loc, NAF::POS, get_clone(it.key()))),
                 std::move(body)));
-            x.done = true;
+            it.value().second = true;
         }
     }
     // }}}3
 }
 
 void Program::check(Logger &log) {
-    for (auto &block : blocks_) {
-        for (auto &stm : block.stms) { stm->check(log); }
+    for (auto const &block : blocks_) {
+        for (auto const &stm : block.second.stms) {
+            stm->check(log);
+        }
     }
     std::unordered_map<Sig, Location> seenSigs;
     for (auto &def : theoryDefs_) {
-        for (auto &atomDef : def.atomDefs()) {
+        for (auto const &atomDef : def.atomDefs()) {
             auto seenSig = seenSigs.emplace(atomDef.sig(), atomDef.loc());
             if (!seenSig.second) {
                 GRINGO_REPORT(log, Warnings::RuntimeError)
@@ -200,71 +228,84 @@ void Program::check(Logger &log) {
 }
 
 void Program::print(std::ostream &out) const {
-    for (auto &def : theoryDefs_) {
+    for (auto const &def : theoryDefs_) {
         out << def << "\n";
     }
-    for (auto &block : blocks_) {
-        for (auto &x : block.addedEdb)          { out << x << "." << "\n"; }
-        for (auto &x : std::get<1>(*block.edb)) { out << x << "." << "\n"; }
-        for (auto &x : block.addedStms)         { out << *x << "\n"; }
-        for (auto &x : block.stms)              { out << *x << "\n"; }
+    for (auto const &block : blocks_) {
+        for (auto const &x : block.second.addedEdb)     { out << x << "." << "\n"; }
+        for (auto const &x : std::get<1>(*block.first)) { out << x << "." << "\n"; }
+        for (auto const &x : block.second.addedStms)    { out << *x << "\n"; }
+        for (auto const &x : block.second.stms)         { out << *x << "\n"; }
     }
-    for (auto &x : stms_) { out << *x << "\n"; }
+    for (auto const &x : stms_) { out << *x << "\n"; }
 }
 
 // Defines atoms that have been seen in earlier steps
 class DummyStatement : public Ground::Statement, private Ground::HeadOccurrence {
 public:
-    DummyStatement(UGTermVec terms, bool normal) : terms_{std::move(terms)}, normal_{normal}  {}
-    bool isNormal() const override { return normal_; }
+    DummyStatement(UGTermVec terms, bool normal)
+    : terms_{std::move(terms)}
+    , normal_{normal}  {}
+
+    bool isNormal() const override {
+        return normal_;
+    }
+
     void analyze(Dep::Node &node, Dep &dep) override {
         for (auto &term : terms_) {
             dep.provides(node, *this, get_clone(term));
         }
     }
-    void startLinearize(bool) override { }
-    void linearize(Context &, bool, Logger &) override { }
-    void enqueue(Ground::Queue &) override { }
+
+    void startLinearize(bool active) override {
+        static_cast<void>(active);
+    }
+
+    void linearize(Context &context, bool positive, Logger &log) override {
+        static_cast<void>(context);
+        static_cast<void>(positive);
+        static_cast<void>(log);
+    }
+
+    void enqueue(Ground::Queue &q) override {
+        static_cast<void>(q);
+    }
+
     void print(std::ostream &out) const override {
         print_comma(out, terms_, ";", [](std::ostream &out, UGTerm const &term) { out << *term; });
         out << ".";
     }
+
 private:
-    void defines(IndexUpdater &, Ground::Instantiator *) override { }
-private:
+    void defines(IndexUpdater &update, Ground::Instantiator *inst) override {
+        static_cast<void>(update);
+        static_cast<void>(inst);
+    }
+
     UGTermVec terms_;
     bool normal_;
 };
 
 Ground::Program Program::toGround(std::set<Sig> const &sigs, DomainData &domains, Logger &log) {
-    HashSet<uint64_t> neg;
-    Ground::Program::ClassicalNegationVec negate;
-    auto gn = [&neg, &negate, &domains](Sig x) {
-        if (neg.insert(std::hash<uint64_t>(), std::equal_to<uint64_t>(), x.rep()).second) {
-            negate.emplace_back(domains.add(x.flipSign()), domains.add(x));
-        }
-    };
     Ground::UStmVec stms;
-    if (!pheads.empty()) { stms.emplace_back(gringo_make_unique<DummyStatement>(std::move(pheads), true)); }
-    if (!nheads.empty()) { stms.emplace_back(gringo_make_unique<DummyStatement>(std::move(nheads), false)); }
+    if (!pheads.empty()) {
+        stms.emplace_back(gringo_make_unique<DummyStatement>(std::move(pheads), true));
+    }
+    if (!nheads.empty()) {
+        stms.emplace_back(gringo_make_unique<DummyStatement>(std::move(nheads), false));
+    }
     stms.emplace_back(make_locatable<Ground::ExternalRule>(Location("#external", 1, 1, "#external", 1, 1)));
     ToGroundArg arg(auxNames_, domains);
     Ground::SEdbVec edb;
-    for (auto &block : blocks_) {
-        if (sigs.find(Sig{block.name.c_str() + 5, numeric_cast<uint32_t>(block.params.size()), false}) != sigs.end()) {
-            for (auto &x : block.edb->second) {
-                auto sig = x.sig();
-                if (sig.sign()) { gn(sig); }
-            }
-            edb.emplace_back(block.edb);
-            for (auto &x : block.stms) {
-                x->getNeg(gn);
+    for (auto const &block : blocks_) {
+        if (sigs.find(Sig{block.second.name.c_str() + 5, numeric_cast<uint32_t>(block.second.params.size()), false}) != sigs.end()) { // NOLINT
+            edb.emplace_back(block.first);
+            for (auto const &x : block.second.stms) {
                 x->toGround(arg, stms);
             }
         }
     }
     for (auto &x : stms_) {
-        x->getNeg(gn);
         x->toGround(arg, stms);
     }
     Ground::Statement::Dep dep;
@@ -274,15 +315,17 @@ Ground::Program Program::toGround(std::set<Sig> const &sigs, DomainData &domains
         node.stm->analyze(node, dep);
     }
     auto ret = dep.analyze();
-    Ground::Program prg(std::move(edb), std::move(std::get<0>(ret)), std::move(negate));
+    Ground::Program prg(std::move(edb), std::move(std::get<0>(ret)));
     pheads = std::move(std::get<1>(ret));
     nheads = std::move(std::get<2>(ret));
-    for (auto &sig : sigs_) {
+    for (auto const &sig : sigs_) {
         domains.add(sig);
     }
     Ground::UndefVec undef;
-    for (auto &x : dep.depend.occs) {
-        for (auto &y : x.second.first->depend) { (*std::get<0>(y)).checkDefined(locs_, sigs_, undef); }
+    for (auto const &x : dep.depend) {
+        for (auto const &y : x.second.first->depend) {
+            (*std::get<0>(y)).checkDefined(locs_, sigs_, undef);
+        }
     }
     std::sort(undef.begin(), undef.end(), [](Ground::UndefVec::value_type const &a, Ground::UndefVec::value_type const &b) { return a.first < b.first; });
     for (auto &x : undef) {
@@ -293,8 +336,6 @@ Ground::Program Program::toGround(std::set<Sig> const &sigs, DomainData &domains
     return prg;
 }
 
-Program::~Program() { }
-
 std::ostream &operator<<(std::ostream &out, Program const &p) {
     p.print(out);
     return out;
