@@ -37,10 +37,10 @@ namespace Gringo { namespace Input {
 
 // {{{1 definition of NongroundProgramBuilder
 
-NongroundProgramBuilder::NongroundProgramBuilder(Context &context, Program &prg, Output::OutputBase &out, Defines &defs, bool rewriteMinimize)
+NongroundProgramBuilder::NongroundProgramBuilder(Context &context, Program &prg, OutputPredicates &output_preds, Defines &defs, bool rewriteMinimize)
 : context_(context)
 , prg_(prg)
-, out(out)
+, output_preds_(output_preds)
 , defs_(defs)
 , rewriteMinimize_(rewriteMinimize)
 { }
@@ -52,12 +52,14 @@ TermUid NongroundProgramBuilder::term(Location const &loc, Symbol val) {
 }
 
 TermUid NongroundProgramBuilder::term(Location const &loc, String name) {
-    if (name == "_") { return terms_.insert(make_locatable<VarTerm>(loc, name, nullptr)); }
-    else {
-        auto &ret(vals_[name]);
-        if (!ret) { ret = std::make_shared<Symbol>(); }
-        return terms_.insert(make_locatable<VarTerm>(loc, name, ret));
+    if (name == "_") {
+        return terms_.insert(make_locatable<VarTerm>(loc, name, nullptr));
     }
+    auto &ret(vals_[name]);
+    if (!ret) {
+        ret = std::make_shared<Symbol>();
+    }
+    return terms_.insert(make_locatable<VarTerm>(loc, name, ret));
 }
 
 TermUid NongroundProgramBuilder::term(Location const &loc, UnOp op, TermUid a) {
@@ -66,12 +68,14 @@ TermUid NongroundProgramBuilder::term(Location const &loc, UnOp op, TermUid a) {
 
 TermUid NongroundProgramBuilder::term(Location const &loc, UnOp op, TermVecUid a) {
     UTermVec vec(termvecs_.erase(a));
-    if (vec.size() == 1) { return terms_.insert(make_locatable<UnOpTerm>(loc, op, std::move(vec.front()))); }
-    else {
-        UTermVec pool;
-        for (auto &terms : vec) { pool.emplace_back(make_locatable<UnOpTerm>(loc, op, std::move(terms))); }
-        return terms_.insert(make_locatable<PoolTerm>(loc, std::move(pool)));
+    if (vec.size() == 1) {
+        return terms_.insert(make_locatable<UnOpTerm>(loc, op, std::move(vec.front())));
     }
+    UTermVec pool;
+    for (auto &terms : vec) {
+        pool.emplace_back(make_locatable<UnOpTerm>(loc, op, std::move(terms)));
+    }
+    return terms_.insert(make_locatable<PoolTerm>(loc, std::move(pool)));
 }
 
 TermUid NongroundProgramBuilder::term(Location const &loc, BinOp op, TermUid a, TermUid b) {
@@ -86,21 +90,27 @@ TermUid NongroundProgramBuilder::term(Location const &loc, String name, TermVecV
     assert(name != "");
     auto create = [&lua, &name, &loc](UTermVec &&vec) -> UTerm {
         // lua terms
-        if (lua) { return make_locatable<LuaTerm>(loc, name, std::move(vec)); }
+        if (lua) {
+            return make_locatable<LuaTerm>(loc, name, std::move(vec));
+        }
         // constant symbols
-        else if (vec.empty()) { return make_locatable<ValTerm>(loc, Symbol::createId(name)); }
+        if (vec.empty()) {
+            return make_locatable<ValTerm>(loc, Symbol::createId(name));
+        }
         // function terms
-        else { return make_locatable<FunctionTerm>(loc, name, std::move(vec)); }
+        return make_locatable<FunctionTerm>(loc, name, std::move(vec));
     };
     TermVecVecs::ValueType vec(termvecvecs_.erase(a));
     // no pooling
-    if (vec.size() == 1) { return terms_.insert(create(std::move(vec.front()))); }
-    // pooling
-    else {
-        UTermVec pool;
-        for (auto &terms : vec) { pool.emplace_back(create(std::move(terms))); }
-        return terms_.insert(make_locatable<PoolTerm>(loc, std::move(pool)));
+    if (vec.size() == 1) {
+        return terms_.insert(create(std::move(vec.front())));
     }
+    // pooling
+    UTermVec pool;
+    for (auto &terms : vec) {
+        pool.emplace_back(create(std::move(terms)));
+    }
+    return terms_.insert(make_locatable<PoolTerm>(loc, std::move(pool)));
 }
 
 TermUid NongroundProgramBuilder::term(Location const &loc, TermVecUid args, bool forceTuple) {
@@ -123,41 +133,6 @@ IdVecUid NongroundProgramBuilder::idvec(IdVecUid uid, Location const &loc, Strin
     return uid;
 }
 
-// {{{2 csp
-
-CSPMulTermUid NongroundProgramBuilder::cspmulterm(Location const &, TermUid coe, TermUid var) {
-    return cspmulterms_.emplace(terms_.erase(var), terms_.erase(coe));
-}
-CSPMulTermUid NongroundProgramBuilder::cspmulterm(Location const &, TermUid coe) {
-    return cspmulterms_.emplace(nullptr, terms_.erase(coe));
-}
-CSPAddTermUid NongroundProgramBuilder::cspaddterm(Location const &loc, CSPAddTermUid a, CSPMulTermUid b, bool add) {
-    if (add) {
-        cspaddterms_[a].append(cspmulterms_.erase(b));
-    }
-    else {
-        CSPMulTerm mul(cspmulterms_.erase(b));
-        mul.coe = make_locatable<UnOpTerm>(loc, UnOp::NEG, std::move(mul.coe));
-        cspaddterms_[a].append(std::move(mul));
-    }
-
-    return a;
-}
-CSPAddTermUid NongroundProgramBuilder::cspaddterm(Location const &, CSPMulTermUid a) {
-    return cspaddterms_.emplace(cspmulterms_.erase(a));
-}
-LitUid NongroundProgramBuilder::csplit(CSPLitUid a) {
-    return lits_.emplace(csplits_.erase(a));
-}
-CSPLitUid NongroundProgramBuilder::csplit(Location const &loc, CSPLitUid a, Relation rel, CSPAddTermUid b) {
-    csplits_[a]->append(rel, cspaddterms_.erase(b));
-    csplits_[a]->loc(csplits_[a]->loc() + loc);
-    return a;
-}
-CSPLitUid NongroundProgramBuilder::csplit(Location const &loc, CSPAddTermUid a, Relation rel, CSPAddTermUid b) {
-    return csplits_.insert(make_locatable<CSPLiteral>(loc, rel, cspaddterms_.erase(a), cspaddterms_.erase(b)));
-}
-
 // {{{2 termvecs
 
 TermVecUid NongroundProgramBuilder::termvec() {
@@ -183,15 +158,26 @@ TermVecVecUid NongroundProgramBuilder::termvecvec(TermVecVecUid uid, TermVecUid
 // {{{2 literals
 
 LitUid NongroundProgramBuilder::boollit(Location const &loc, bool type) {
-    return rellit(loc, type ? Relation::EQ : Relation::NEQ, term(loc, Symbol::createNum(0)), term(loc, Symbol::createNum(0)));
+    return rellit(loc, NAF::POS, term(loc, Symbol::createNum(0)), rellitvec(loc, type ? Relation::EQ : Relation::NEQ, term(loc, Symbol::createNum(0))));
 }
 
 LitUid NongroundProgramBuilder::predlit(Location const &loc, NAF naf, TermUid term) {
     return lits_.insert(make_locatable<PredicateLiteral>(loc, naf, terms_.erase(term)));
 }
 
-LitUid NongroundProgramBuilder::rellit(Location const &loc, Relation rel, TermUid termUidLeft, TermUid termUidRight) {
-    return lits_.insert(make_locatable<RelationLiteral>(loc, rel, terms_.erase(termUidLeft), terms_.erase(termUidRight)));
+RelLitVecUid NongroundProgramBuilder::rellitvec(Location const &loc, Relation rel, TermUid termUidLeft) {
+    auto uid = rellitvecs_.emplace();
+    rellitvecs_[uid].emplace_back(rel, terms_.erase(termUidLeft));
+    return uid;
+}
+
+RelLitVecUid NongroundProgramBuilder::rellitvec(Location const &loc, RelLitVecUid vecUidLeft, Relation rel, TermUid termUidRight) {
+    rellitvecs_[vecUidLeft].emplace_back(rel, terms_.erase(termUidRight));
+    return vecUidLeft;
+}
+
+LitUid NongroundProgramBuilder::rellit(Location const &loc, NAF naf, TermUid termUidLeft, RelLitVecUid vecUidRight) {
+    return lits_.insert(make_locatable<RelationLiteral>(loc, naf, terms_.erase(termUidLeft), rellitvecs_.erase(vecUidRight)));
 }
 
 // {{{2 literal vectors
@@ -273,12 +259,6 @@ BdLitVecUid NongroundProgramBuilder::conjunction(BdLitVecUid body, Location cons
     return body;
 }
 
-BdLitVecUid NongroundProgramBuilder::disjoint(BdLitVecUid body, Location const &loc, NAF naf, CSPElemVecUid elem) {
-    bodies_[body].push_back(make_locatable<DisjointAggregate>(loc, naf, cspelems_.erase(elem)));
-    return body;
-}
-
-
 // {{{2 rule heads
 
 HdLitUid NongroundProgramBuilder::headlit(LitUid lit) {
@@ -297,17 +277,6 @@ HdLitUid NongroundProgramBuilder::disjunction(Location const &loc, CondLitVecUid
     return heads_.insert(make_locatable<Disjunction>(loc, condlitvecs_.erase(condlitvec)));
 }
 
-// {{{2 csp constraint elements
-
-CSPElemVecUid NongroundProgramBuilder::cspelemvec() {
-    return cspelems_.emplace();
-}
-
-CSPElemVecUid NongroundProgramBuilder::cspelemvec(CSPElemVecUid uid, Location const &loc, TermVecUid termvec, CSPAddTermUid addterm, LitVecUid litvec) {
-    cspelems_[uid].emplace_back(loc, termvecs_.erase(termvec), cspaddterms_.erase(addterm), litvecs_.erase(litvec));
-    return uid;
-}
-
 // {{{2 statements
 
 void NongroundProgramBuilder::rule(Location const &loc, HdLitUid head) {
@@ -328,31 +297,28 @@ void NongroundProgramBuilder::optimize(Location const &loc, TermUid weight, Term
         termvec(argsUid, term(loc, cond, true));
         auto predUid = predlit(loc, NAF::POS, term(loc, "_criteria", termvecvec(termvecvec(), argsUid), false));
         rule(loc, headlit(predUid), body);
-        out.outPredsForce.emplace_back(loc, Sig("_criteria", 3, false), false);
+        output_preds_.add(loc, Sig("_criteria", 3, false), true);
     }
     else {
         prg_.add(make_locatable<Statement>(loc, make_locatable<MinimizeHeadLiteral>(loc, terms_.erase(weight), terms_.erase(priority), termvecs_.erase(cond)), bodies_.erase(body)));
     }
 }
 
-void NongroundProgramBuilder::showsig(Location const &loc, Sig sig, bool csp) {
-    out.outPreds.emplace_back(loc, sig, csp);
+void NongroundProgramBuilder::showsig(Location const &loc, Sig sig) {
+    output_preds_.add(loc, sig, false);
 }
 
-void NongroundProgramBuilder::defined(Location const &, Sig sig) {
+void NongroundProgramBuilder::defined(Location const &loc, Sig sig) {
+    static_cast<void>(loc);
     prg_.addInput(sig);
 }
 
-void NongroundProgramBuilder::show(Location const &loc, TermUid t, BdLitVecUid body, bool csp) {
-    prg_.add(make_locatable<Statement>(loc, make_locatable<ShowHeadLiteral>(loc, terms_.erase(t), csp), bodies_.erase(body)));
+void NongroundProgramBuilder::show(Location const &loc, TermUid t, BdLitVecUid body) {
+    prg_.add(make_locatable<Statement>(loc, make_locatable<ShowHeadLiteral>(loc, terms_.erase(t)), bodies_.erase(body)));
 }
 
-void NongroundProgramBuilder::lua(Location const &loc, String code) {
-    context_.exec(ScriptType::Lua, loc, code);
-}
-
-void NongroundProgramBuilder::python(Location const &loc, String code) {
-    context_.exec(ScriptType::Python, loc, code);
+void NongroundProgramBuilder::script(Location const &loc, String type, String code) {
+    context_.exec(type, loc, code);
 }
 
 void NongroundProgramBuilder::block(Location const &loc, String name, IdVecUid args) {
@@ -430,7 +396,9 @@ TheoryTermUid NongroundProgramBuilder::theorytermvalue(Location const &loc, Symb
 
 TheoryTermUid NongroundProgramBuilder::theorytermvar(Location const &loc, String var) {
     auto &ret(vals_[var]);
-    if (!ret) { ret = std::make_shared<Symbol>(); }
+    if (!ret) {
+        ret = std::make_shared<Symbol>();
+    }
     return theoryTerms_.emplace(gringo_make_unique<Output::TermTheoryTerm>(make_locatable<VarTerm>(loc, var, ret)));
 }
 
@@ -457,11 +425,15 @@ TheoryOpVecUid NongroundProgramBuilder::theoryops(TheoryOpVecUid ops, String op)
 TheoryOptermVecUid NongroundProgramBuilder::theoryopterms() {
     return theoryOptermVecs_.emplace();
 }
-TheoryOptermVecUid NongroundProgramBuilder::theoryopterms(TheoryOptermVecUid opterms, Location const &, TheoryOptermUid opterm) {
+
+TheoryOptermVecUid NongroundProgramBuilder::theoryopterms(TheoryOptermVecUid opterms, Location const &loc, TheoryOptermUid opterm) {
+    static_cast<void>(loc);
     theoryOptermVecs_[opterms].emplace_back(gringo_make_unique<Output::RawTheoryTerm>(theoryOpterms_.erase(opterm)));
     return opterms;
 }
-TheoryOptermVecUid NongroundProgramBuilder::theoryopterms(Location const &, TheoryOptermUid opterm, TheoryOptermVecUid opterms) {
+
+TheoryOptermVecUid NongroundProgramBuilder::theoryopterms(Location const &loc, TheoryOptermUid opterm, TheoryOptermVecUid opterms) {
+    static_cast<void>(loc);
     theoryOptermVecs_[opterms].insert(theoryOptermVecs_[opterms].begin(), gringo_make_unique<Output::RawTheoryTerm>(theoryOpterms_.erase(opterm)));
     return opterms;
 }
@@ -469,6 +441,7 @@ TheoryOptermVecUid NongroundProgramBuilder::theoryopterms(Location const &, Theo
 TheoryElemVecUid NongroundProgramBuilder::theoryelems() {
     return theoryElems_.emplace();
 }
+
 TheoryElemVecUid NongroundProgramBuilder::theoryelems(TheoryElemVecUid elems, TheoryOptermVecUid opterms, LitVecUid cond) {
     theoryElems_[elems].emplace_back(theoryOptermVecs_.erase(opterms), litvecs_.erase(cond));
     return elems;
@@ -477,7 +450,9 @@ TheoryElemVecUid NongroundProgramBuilder::theoryelems(TheoryElemVecUid elems, Th
 TheoryAtomUid NongroundProgramBuilder::theoryatom(TermUid term, TheoryElemVecUid elems) {
     return theoryAtoms_.emplace(terms_.erase(term), theoryElems_.erase(elems));
 }
-TheoryAtomUid NongroundProgramBuilder::theoryatom(TermUid term, TheoryElemVecUid elems, String op, Location const &, TheoryOptermUid opterm) {
+
+TheoryAtomUid NongroundProgramBuilder::theoryatom(TermUid term, TheoryElemVecUid elems, String op, Location const &loc, TheoryOptermUid opterm) {
+    static_cast<void>(loc);
     return theoryAtoms_.emplace(terms_.erase(term), theoryElems_.erase(elems), op, gringo_make_unique<Output::RawTheoryTerm>(theoryOpterms_.erase(opterm)));
 }
 
@@ -510,7 +485,7 @@ TheoryTermDefUid NongroundProgramBuilder::theorytermdef(Location const &loc, Str
     for (auto &opDef : theoryOpDefVecs_.erase(defs)) {
         def.addOpDef(std::move(opDef), log);
     }
-    return theoryTermDefs_.emplace(std::move(def));
+    return theoryTermDefs_.insert(std::move(def));
 }
 
 TheoryAtomDefUid NongroundProgramBuilder::theoryatomdef(Location const &loc, String name, unsigned arity, String termDef, TheoryAtomType type) {
@@ -549,7 +524,7 @@ void NongroundProgramBuilder::theorydef(Location const &loc, String name, Theory
 
 // }}}2
 
-NongroundProgramBuilder::~NongroundProgramBuilder() { }
+NongroundProgramBuilder::~NongroundProgramBuilder() noexcept = default;
 
 // }}}1
 
