@@ -51,14 +51,6 @@ public:
     virtual TermUid term(Location const &loc, TermVecUid args, bool forceTuple) override;
     virtual TermUid pool(Location const &loc, TermVecUid args) override;
     // }}}
-    // {{{ csp
-    virtual CSPMulTermUid cspmulterm(Location const &loc, TermUid coe, TermUid var) override;
-    virtual CSPMulTermUid cspmulterm(Location const &loc, TermUid coe) override;
-    virtual CSPAddTermUid cspaddterm(Location const &loc, CSPAddTermUid a, CSPMulTermUid b, bool add) override;
-    virtual CSPAddTermUid cspaddterm(Location const &loc, CSPMulTermUid a) override;
-    virtual LitUid csplit(CSPLitUid a) override;
-    virtual CSPLitUid csplit(Location const &loc, CSPLitUid a, Relation rel, CSPAddTermUid b) override;
-    virtual CSPLitUid csplit(Location const &loc, CSPAddTermUid a, Relation rel, CSPAddTermUid b) override;
     // }}}
     // {{{ id vectors
     virtual IdVecUid idvec() override;
@@ -75,7 +67,9 @@ public:
     // {{{ literals
     virtual LitUid boollit(Location const &loc, bool type) override;
     virtual LitUid predlit(Location const &loc, NAF naf, TermUid termUid) override;
-    virtual LitUid rellit(Location const &loc, Relation rel, TermUid termUidLeft, TermUid termUidRight) override;
+    virtual RelLitVecUid rellitvec(Location const &loc, Relation rel, TermUid termUidLeft) override;
+    virtual RelLitVecUid rellitvec(Location const &loc, RelLitVecUid vecUidLeft, Relation rel, TermUid termUidRight) override;
+    virtual LitUid rellit(Location const &loc, NAF naf, TermUid termUidLeft, RelLitVecUid vecUidRight) override;
     // }}}
     // {{{ literal vectors
     virtual LitVecUid litvec() override;
@@ -111,22 +105,16 @@ public:
     virtual BdLitVecUid bodyaggr(BdLitVecUid body, Location const &loc, NAF naf, AggregateFunction fun, BoundVecUid bounds, BdAggrElemVecUid bodyaggrelemvec) override;
     virtual BdLitVecUid bodyaggr(BdLitVecUid body, Location const &loc, NAF naf, AggregateFunction fun, BoundVecUid bounds, CondLitVecUid bodyaggrelemvec) override;
     virtual BdLitVecUid conjunction(BdLitVecUid body, Location const &loc, LitUid head, LitVecUid litvec) override;
-    virtual BdLitVecUid disjoint(BdLitVecUid body, Location const &loc, NAF naf, CSPElemVecUid elem) override;
-    // }}}
-    // {{{ csp constraint elements
-    virtual CSPElemVecUid cspelemvec() override;
-    virtual CSPElemVecUid cspelemvec(CSPElemVecUid uid, Location const &loc, TermVecUid termvec, CSPAddTermUid addterm, LitVecUid litvec) override;
     // }}}
     // {{{ statements
     virtual void rule(Location const &loc, HdLitUid head) override;
     virtual void rule(Location const &loc, HdLitUid head, BdLitVecUid body) override;
     virtual void define(Location const &loc, String name, TermUid value, bool defaultDef, Logger &log) override;
     virtual void optimize(Location const &loc, TermUid weight, TermUid priority, TermVecUid cond, BdLitVecUid body) override;
-    virtual void showsig(Location const &loc, Sig sig, bool csp) override;
+    virtual void showsig(Location const &loc, Sig sig) override;
     virtual void defined(Location const &loc, Sig sig) override;
-    virtual void show(Location const &loc, TermUid t, BdLitVecUid body, bool csp) override;
-    virtual void python(Location const &loc, String code) override;
-    virtual void lua(Location const &loc, String code) override;
+    virtual void show(Location const &loc, TermUid t, BdLitVecUid body) override;
+    virtual void script(Location const &loc, String type, String code) override;
     virtual void block(Location const &loc, String name, IdVecUid args) override;
     virtual void external(Location const &loc, TermUid head, BdLitVecUid body, TermUid type) override;
     virtual void edge(Location const &loc, TermVecVecUid edges, BdLitVecUid body) override;
@@ -197,15 +185,12 @@ private:
     using TermVecVecs = Indexed<StringVecVec, TermVecVecUid>;
     using Lits = Indexed<std::string, LitUid>;
     using LitVecs = Indexed<StringVec, LitVecUid>;
+    using RelLitVecs = Indexed<StringVec, RelLitVecUid>;
     using BodyAggrElemVecs = Indexed<StringVec, BdAggrElemVecUid>;
     using CondLitVecs = Indexed<StringVec, CondLitVecUid>;
     using HeadAggrElemVecs = Indexed<StringVec, HdAggrElemVecUid>;
     using Bodies = Indexed<StringVec, BdLitVecUid>;
-    using CSPElems = Indexed<StringVec, CSPElemVecUid>;
     using Heads = Indexed<std::string, HdLitUid>;
-    using CSPAddTerms = Indexed<std::string, CSPAddTermUid>;
-    using CSPMulTerms = Indexed<std::string, CSPMulTermUid>;
-    using CSPLits = Indexed<std::string, CSPLitUid>;
     using Bounds = Indexed<StringPairVec, BoundVecUid>;
     using Statements = std::vector<std::string>;
 
@@ -237,15 +222,12 @@ private:
     TermVecVecs termvecvecs_;
     Lits lits_;
     LitVecs litvecs_;
+    RelLitVecs rellitvecs_;
     BodyAggrElemVecs bodyaggrelemvecs_;
     CondLitVecs condlitvecs_;
     HeadAggrElemVecs headaggrelemvecs_;
     Bodies bodies_;
     Heads heads_;
-    CSPAddTerms cspaddterms_;
-    CSPMulTerms cspmulterms_;
-    CSPLits csplits_;
-    CSPElems cspelems_;
     Bounds bounds_;
     Statements statements_;
     std::stringstream current_;
@@ -346,35 +328,6 @@ TermUid TestNongroundProgramBuilder::pool(Location const &, TermVecUid args) {
     return terms_.emplace(str());
 }
 
-// }}}
-// {{{ csp
-CSPMulTermUid TestNongroundProgramBuilder::cspmulterm(Location const &, TermUid coe, TermUid var) {
-    current_ << terms_.erase(coe) << "$*$" << terms_.erase(var);
-    return cspmulterms_.emplace(str());
-}
-CSPMulTermUid TestNongroundProgramBuilder::cspmulterm(Location const &, TermUid coe) {
-    current_ << terms_.erase(coe);
-    return cspmulterms_.emplace(str());
-}
-CSPAddTermUid TestNongroundProgramBuilder::cspaddterm(Location const &, CSPAddTermUid a, CSPMulTermUid b, bool add) {
-    cspaddterms_[a] += (add ? "$+" : "$-") + cspmulterms_.erase(b);
-    return a;
-}
-CSPAddTermUid TestNongroundProgramBuilder::cspaddterm(Location const &, CSPMulTermUid a) {
-    current_ << cspmulterms_.erase(a);
-    return cspaddterms_.emplace(str());
-}
-LitUid TestNongroundProgramBuilder::csplit(CSPLitUid a) {
-    return lits_.emplace(csplits_.erase(a));
-}
-CSPLitUid TestNongroundProgramBuilder::csplit(Location const &, CSPLitUid a, Relation rel, CSPAddTermUid b) {
-    csplits_[a] += "$" + IO::to_string(rel) + cspaddterms_.erase(b);
-    return a;
-}
-CSPLitUid TestNongroundProgramBuilder::csplit(Location const &, CSPAddTermUid a, Relation rel, CSPAddTermUid b) {
-    current_ << cspaddterms_.erase(a) << "$" << rel << cspaddterms_.erase(b);
-    return csplits_.emplace(str());
-}
 // }}}
 // {{{ term vectors
 
@@ -412,8 +365,21 @@ LitUid TestNongroundProgramBuilder::predlit(Location const &, NAF naf, TermUid t
     return lits_.emplace(str());
 }
 
-LitUid TestNongroundProgramBuilder::rellit(Location const &, Relation rel, TermUid termUidLeft, TermUid termUidRight) {
-    current_ << terms_.erase(termUidLeft) << rel << terms_.erase(termUidRight);
+RelLitVecUid TestNongroundProgramBuilder::rellitvec(Location const &loc, Relation rel, TermUid termUidLeft) {
+    auto uid = rellitvecs_.emplace();
+    return rellitvec(loc, uid, rel, termUidLeft);
+}
+
+RelLitVecUid TestNongroundProgramBuilder::rellitvec(Location const &loc, RelLitVecUid vecUidLeft, Relation rel, TermUid termUidRight) {
+    current_<< rel << terms_.erase(termUidRight);
+    rellitvecs_[vecUidLeft].emplace_back(str());
+    return vecUidLeft;
+}
+
+LitUid TestNongroundProgramBuilder::rellit(Location const &loc, NAF naf, TermUid termUidLeft, RelLitVecUid vecUidRight) {
+    print(naf);
+    current_ << terms_.erase(termUidLeft);
+    print(rellitvecs_.erase(vecUidRight), "");
     return lits_.emplace(str());
 }
 
@@ -556,29 +522,6 @@ BdLitVecUid TestNongroundProgramBuilder::conjunction(BdLitVecUid uid, Location c
     return uid;
 }
 
-BdLitVecUid TestNongroundProgramBuilder::disjoint(BdLitVecUid body, Location const &, NAF naf, CSPElemVecUid elem) {
-    current_ << naf << "#disjoint{";
-    print(cspelems_.erase(elem), ",");
-    current_ << "}";
-    bodies_[body].emplace_back(str());
-    return body;
-}
-
-// }}}
-// {{{ csp constraint elements
-
-CSPElemVecUid TestNongroundProgramBuilder::cspelemvec() {
-    return cspelems_.emplace();
-}
-
-CSPElemVecUid TestNongroundProgramBuilder::cspelemvec(CSPElemVecUid uid, Location const &, TermVecUid termvec, CSPAddTermUid addterm, LitVecUid litvec) {
-    print(termvecs_.erase(termvec), ",");
-    current_ << ":" << cspaddterms_.erase(addterm) << ":";
-    print(litvecs_.erase(litvec), ",");
-    cspelems_[uid].emplace_back(str());
-    return uid;
-}
-
 // }}}
 // {{{ statements
 
@@ -618,8 +561,8 @@ void TestNongroundProgramBuilder::optimize(Location const &, TermUid weight, Ter
     statements_.emplace_back(str());
 }
 
-void TestNongroundProgramBuilder::showsig(Location const &, Sig sig, bool csp) {
-    current_ << "#showsig " << (csp ? "$" : "") << sig << ".";
+void TestNongroundProgramBuilder::showsig(Location const &, Sig sig) {
+    current_ << "#showsig " << sig << ".";
     statements_.emplace_back(str());
 }
 
@@ -628,8 +571,8 @@ void TestNongroundProgramBuilder::defined(Location const &, Sig sig) {
     statements_.emplace_back(str());
 }
 
-void TestNongroundProgramBuilder::show(Location const &, TermUid t, BdLitVecUid bodyuid, bool csp) {
-    current_ << "#show " << (csp ? "$" : "") << terms_.erase(t);
+void TestNongroundProgramBuilder::show(Location const &, TermUid t, BdLitVecUid bodyuid) {
+    current_ << "#show " << terms_.erase(t);
     StringVec body(bodies_.erase(bodyuid));
     if (!body.empty()) {
         current_ << ":";
@@ -639,15 +582,11 @@ void TestNongroundProgramBuilder::show(Location const &, TermUid t, BdLitVecUid
     statements_.emplace_back(str());
 }
 
-void TestNongroundProgramBuilder::python(Location const &, String code) {
-    current_ << code << ".";
+void TestNongroundProgramBuilder::script(Location const &, String type, String code) {
+    current_ << "#script(" << type << ")\n" << code << "\n#end.";
     statements_.emplace_back(str());
 }
 
-void TestNongroundProgramBuilder::lua(Location const &, String code) {
-    current_ << code << ".";
-    statements_.emplace_back(str());
-}
 void TestNongroundProgramBuilder::block(Location const &, String name, IdVecUid args) {
     current_ << "#program " << name << "(";
     print(idvecs_.erase(args), ",");
@@ -944,7 +883,8 @@ std::string parse(std::string const &str) {
     Gringo::Test::TestGringoModule log;
     TestNongroundProgramBuilder pb;
     bool incmode;
-    NonGroundParser ngp(pb, incmode);
+    NullBackend bck;
+    NonGroundParser ngp(pb, bck, incmode);
     ngp.pushStream("-", std::unique_ptr<std::istream>(new std::stringstream(str)), log);
     ngp.parse(log);
     return pb.toString();
@@ -1056,6 +996,10 @@ TEST_CASE("input-nongroundprogrambuilder", "[input]") {
         REQUIRE("#program base().\n#not #not p." == parse("not not p."));
         // term cmp term
         REQUIRE("#program base().\n(1+2)!=3." == parse("1+2!=3."));
+        REQUIRE("#program base().\n(1+2)<3<4." == parse("1+2<3<4."));
+        REQUIRE("#program base().\n(1+2)<3<4<=5." == parse("1+2<3<4<=5."));
+        REQUIRE("#program base().\n#not (1+2)<3<4<=5." == parse("not 1+2<3<4<=5."));
+        REQUIRE("#program base().\n#not #not (1+2)<3<4<=5." == parse("not not 1+2<3<4<=5."));
     }
 
     SECTION("bdaggr") {
@@ -1144,23 +1088,25 @@ TEST_CASE("input-nongroundprogrambuilder", "[input]") {
         REQUIRE("#program base().\na:b." == parse("a:b."));
         REQUIRE("#program base().\na:b,c." == parse("a:b,c."));
         // literal COMMA disjunctionsep literal optcondition
-        REQUIRE("#program base().\nb:;c:;a:." == parse("a,b,c."));
-        REQUIRE("#program base().\nb:;c:;d:;a:." == parse("a,b;c;d."));
-        REQUIRE("#program base().\nb:;c:d,e;a:." == parse("a,b,c:d,e."));
-        REQUIRE("#program base().\nb:d,e;c:;a:." == parse("a,b:d,e;c."));
+        REQUIRE("#program base().\na:;b:;c:." == parse("a,b,c."));
+        REQUIRE("#program base().\na:;b:;c:;d:." == parse("a,b;c;d."));
+        REQUIRE("#program base().\na:;b:;c:d,e." == parse("a,b,c:d,e."));
+        REQUIRE("#program base().\na:;b:d,e;c:." == parse("a,b:d,e;c."));
         // literal SEM disjunctionsep literal optcondition
-        REQUIRE("#program base().\nb:;c:;a:." == parse("a;b,c."));
-        REQUIRE("#program base().\nb:;c:;d:;a:." == parse("a;b;c;d."));
-        REQUIRE("#program base().\nb:;c:d,e;a:." == parse("a;b,c:d,e."));
-        REQUIRE("#program base().\nb:d,e;c:;a:." == parse("a;b:d,e;c."));
+        REQUIRE("#program base().\na:;b:;c:." == parse("a;b,c."));
+        REQUIRE("#program base().\na:;b:;c:;d:." == parse("a;b;c;d."));
+        REQUIRE("#program base().\na:;b:;c:d,e." == parse("a;b,c:d,e."));
+        REQUIRE("#program base().\na:;b:d,e;c:." == parse("a;b:d,e;c."));
         // literal COLON litvec SEM disjunctionsep literal optcondition
-        REQUIRE("#program base().\nc:;a:." == parse("a;c."));
-        REQUIRE("#program base().\nc:;a:x." == parse("a:x;c."));
-        REQUIRE("#program base().\nc:;a:x,y." == parse("a:x,y;c."));
-        REQUIRE("#program base().\nb:;c:;a:x,y." == parse("a:x,y;b,c."));
-        REQUIRE("#program base().\nb:;c:;d:;a:x,y." == parse("a:x,y;b;c;d."));
-        REQUIRE("#program base().\nb:;c:d,e;a:x,y." == parse("a:x,y;b,c:d,e."));
-        REQUIRE("#program base().\nb:d,e;c:;a:x,y." == parse("a:x,y;b:d,e;c."));
+        REQUIRE("#program base().\na:;c:." == parse("a;c."));
+        REQUIRE("#program base().\na:x;c:." == parse("a:x;c."));
+        REQUIRE("#program base().\na:x,y;c:." == parse("a:x,y;c."));
+        REQUIRE("#program base().\na:x,y;b:;c:." == parse("a:x,y;b,c."));
+        REQUIRE("#program base().\na:x,y;b:;c:;d:." == parse("a:x,y;b;c;d."));
+        REQUIRE("#program base().\na:x,y;b:;c:d,e." == parse("a:x,y;b,c:d,e."));
+        REQUIRE("#program base().\na:x,y;b:d,e;c:." == parse("a:x,y;b:d,e;c."));
+        // empty condition
+        REQUIRE("#program base().\na:;b:;c:." == parse("a:;b:;c:."));
     }
 
     SECTION("external") {
@@ -1190,7 +1136,7 @@ TEST_CASE("input-nongroundprogrambuilder", "[input]") {
         REQUIRE("#program base().\n#count{}." == parse("{}."));
         REQUIRE("#program base().\nnott<=#count{}." == parse("nott{}."));
         // disjunction
-        REQUIRE("#program base().\nb:;a:." == parse("a,b."));
+        REQUIRE("#program base().\na:;b:." == parse("a,b."));
         // rules
         REQUIRE("#program base().\na." == parse("a."));
         REQUIRE("#program base().\na:-b." == parse("a:-b."));
@@ -1274,12 +1220,6 @@ TEST_CASE("input-nongroundprogrambuilder", "[input]") {
         REQUIRE("#program base().\na.\nb.\n#program base().\nc.\nd." == parse("a.\n#include \"test_include.lp\".\nc.\nd.\n"));
     }
 
-    SECTION("csp") {
-        REQUIRE("#program base().\n1$<=1$*$x$<=10." == parse("1 $<= $x $<= 10."));
-        REQUIRE("#program base().\n1$*$a$<=X$*$x$+8$*$X$<=10:-p(X)." == parse("1 $* $a $<= $x $* X $+ 8 $* $X $<= 10 :- p(X)."));
-        REQUIRE("#program base().\n#false:-a;b;not #disjoint{a:1$*$a:a,a,b:1$*$b:b}." == parse("#disjoint{a:$a:a,a; b:$b:b} :- a, b."));
-    }
-
     SECTION("edge") {
         REQUIRE("#program base().\n#edge(a,b)." == parse("#edge (a, b)."));
         REQUIRE("#program base().\n#edge(a,b;e,f)." == parse("#edge (a, b;e,f):."));
