@@ -42,8 +42,8 @@ namespace {
 struct Grounder {
     Grounder()
     : out(td, {}, oss)
-    , pb( context, p, out, d )
-    , ngp( pb, incmode )
+    , pb( context, p, out.outPreds, d )
+    , ngp( pb, bck, incmode )
     { }
     Gringo::Test::TestGringoModule module;
     Potassco::TheoryData td;
@@ -52,9 +52,10 @@ struct Grounder {
     Gringo::Test::TestContext context;
     Program p;
     Defines d;
+    NullBackend bck;
     NongroundProgramBuilder pb;
     NonGroundParser ngp;
-    bool incmode;
+    bool incmode{false};
 
 };
 
@@ -98,7 +99,7 @@ TEST_CASE("input-program", "[input]") {
         REQUIRE("p(1):-q(3).p(2):-q(3).p(1):-q(4).p(2):-q(4)." == rewrite(parse("p(1;2):-q(3;4).")));
         REQUIRE("p((X+Y)):-q(#Arith0);#Arith0=(X+Y)." == rewrite(parse("p(X+Y):-q(X+Y).")));
         REQUIRE("#Arith0<=#count{(X+Y):q((X+Y)):r(#Arith0),s(#Arith1),#Arith1=(A+B)}:-t(#Arith0);#Arith0=(X+Y);1<=#count{(X+Y):u(#Arith0),v(#Arith2),#Arith2=(A+B)}." == rewrite(parse("X+Y#count{X+Y:q(X+Y):r(X+Y),s(A+B)}:-t(X+Y),1#count{X+Y:u(X+Y),v(A+B)}.")));
-        REQUIRE("p(#Range0):-q(#Range1);#range(#Range1,A,B);#range(#Range0,X,Y)." == rewrite(parse("p(X..Y):-q(A..B).")));
+        REQUIRE("p((#Range0+0)):-q((#Range1+0));#range(#Range1,A,B);#range(#Range0,X,Y)." == rewrite(parse("p(X..Y):-q(A..B).")));
         REQUIRE("p(1):-q.p(2):-q.p(3):-q.p:-q." == rewrite(parse("p(1;2;3;):-q.")));
         REQUIRE("p(Z):-p(A,B);Y=#count{0,q(B):q(B)};X=#count{0,q(A):q(A)};Z=#count{0,r(X,Y):r(X,Y)}." == rewrite(parse("p(Z):-p(A,B),X={q(A)},Y={q(B)},Z={r(X,Y)}.")));
         REQUIRE("p(Z):-Z=#count{0,p(X):p(X)};Z>0." == rewrite(parse("p(Z):-Z={p(X)},Z>0.")));
@@ -107,12 +108,160 @@ TEST_CASE("input-program", "[input]") {
         REQUIRE("#project p(X):-[p(X)].#project p(Y):-[p(Y)]." == rewrite(parse("#project p(X;Y).")));
         REQUIRE("#edge(a,b).#edge(c,d)." == rewrite(parse("#edge (a,b;c,d).")));
         REQUIRE("#edge((X+X),b)." == rewrite(parse("#edge (X+X,b).")));
-        REQUIRE("#edge(#Range0,b):-#range(#Range0,1,10)." == rewrite(parse("#edge (1..10,b).")));
+        REQUIRE("#edge((#Range0+0),b):-#range(#Range0,1,10)." == rewrite(parse("#edge (1..10,b).")));
         REQUIRE("#heuristic a(#Arith0)[1@0,sign]:-[a(#Arith0)];#Arith0=(X+X)." == rewrite(parse("#heuristic a(X+X). [1@0,sign]")));
         REQUIRE("#heuristic a[(X+X)@2,sign]:-[a]." == rewrite(parse("#heuristic a. [X+X@2,sign]")));
-        REQUIRE("#heuristic a(#Range0)[2@0,sign]:-[a(#Range0)];#range(#Range0,1,2)." == rewrite(parse("#heuristic a(1..2). [2,sign]")));
+        REQUIRE("#heuristic a((#Range0+0))[2@0,sign]:-[a((#Range0+0))];#range(#Range0,1,2)." == rewrite(parse("#heuristic a(1..2). [2,sign]")));
         REQUIRE("#theory x{node{};&edge/1:node,head}.#false:-p(Z);not &edge((Z+Z)){(z),(Y): p(Y,#Arith1),#Arith1=(Y+Y)}." == rewrite(parse("#theory x{ node{}; &edge/1: node, head }.&edge(Z+Z) { z, Y : p(Y,Y+Y)} :- p(Z).")));
-        REQUIRE("#theory x{node{};&edge/1:node,head}.#false:-p(Z);#range(#Range0,Z,Z);not &edge(#Range0){(z),(Y): p(Y,#Range1),#range(#Range1,Y,Y)}." == rewrite(parse("#theory x{ node{}; &edge/1: node, head }.&edge(Z..Z) { z, Y : p(Y,Y..Y)} :- p(Z).")));
+        REQUIRE("#theory x{node{};&edge/1:node,head}.#false:-p(Z);#range(#Range0,Z,Z);not &edge((#Range0+0)){(z),(Y): p(Y,(#Range1+0)),#range(#Range1,Y,Y)}." == rewrite(parse("#theory x{ node{}; &edge/1: node, head }.&edge(Z..Z) { z, Y : p(Y,Y..Y)} :- p(Z).")));
+    }
+
+    SECTION("rewrite-comparisons") {
+        // head
+        REQUIRE("#void:-1>=2;q.#void:-2>=3;q." == rewrite(parse("1<2<3:-q.")));
+        REQUIRE("#void:-2<3;1<2;q." == rewrite(parse("not 1<2<3:-q.")));
+        REQUIRE("#void::;#void:::-5<6;4<5;2<3;1<2;q." == rewrite(parse("not 1<2<3|not 4<5<6:-q.")));
+        REQUIRE("#void:-#range(#Range0,1,1);#range(#Range0,1,2);q;1>=(#Range0+0)."
+                "#void:-#range(#Range0,1,0);#range(#Range0,1,2);q;(#Range0+0)>=3." == rewrite(parse("1<1..2<3:-q.")));
+        REQUIRE("#void:-#range(#Range0,2,2);#range(#Range0,1,2);q;1<(#Range0+0);(#Range0+0)<3." == rewrite(parse("not 1<1..2<3:-q.")));
+        REQUIRE("#void:-1>=2;q.#void:-2>=4;q.#void:-1>=3;q.#void:-3>=4;q." == rewrite(parse("1<(2;3)<4:-q.")));
+        REQUIRE("#void:-2<4;1<2;q.#void:-3<4;1<3;q." == rewrite(parse("not 1<(2;3)<4:-q.")));
+        // body
+        REQUIRE("q:-2<3;1<2." == rewrite(parse("q :- 1<2<3.")));
+        REQUIRE("q:-1>=2."
+                "q:-2>=3." == rewrite(parse("q :- not 1<2<3.")));
+        REQUIRE("q:-#range(#Range0,2,2);#range(#Range0,1,2);1<(#Range0+0);(#Range0+0)<3." == rewrite(parse("q :- 1<1..2<3.")));
+        REQUIRE("q:-#range(#Range0,1,1);#range(#Range0,1,2);1>=(#Range0+0)."
+                "q:-#range(#Range0,1,0);#range(#Range0,1,2);(#Range0+0)>=3." == rewrite(parse("q :- not 1<1..2<3.")));
+        REQUIRE("q:-2<4;1<2."
+                "q:-3<4;1<3." == rewrite(parse("q :- 1<(2;3)<4.")));
+        REQUIRE("q:-1>=2."
+                "q:-2>=4."
+                "q:-1>=3."
+                "q:-3>=4." == rewrite(parse("q :- not 1<(2;3)<4.")));
+        // disjunctions
+        REQUIRE("#void:1>=2&#void:2>=3:;"
+                "#void:4>=5&#void:5>=6::-q." == rewrite(parse("1<2<3|4<5<6:-q.")));
+        REQUIRE("#void:#range(#Range0,2,3),1>=(#Range0+0)&#void:#range(#Range0,2,3),(#Range0+0)>=4:;"
+                "#void:4>=5&#void:5>=7&#void:4>=6&#void:6>=7::-q." == rewrite(parse("1<(2..3)<4|4<(5;6)<7:-q.")));
+        REQUIRE("#void::;#void:::-5<6;4<5;2<3;1<2;q." == rewrite(parse("not 1<2<3|not 4<5<6:-q.")));
+        REQUIRE("#void:1<2,2<3:b;"
+                "#void:4<5,5<6:c:-q." == rewrite(parse("not 1<2<3:b|not 4<5<6:c:-q.")));
+        REQUIRE("#void::;#void:4<5,5<7&#void:4<6,6<7:"
+                ":-#range(#Range0,2,3);q;1<(#Range0+0);(#Range0+0)<4." == rewrite(parse("not 1<(2..3)<4|not 4<(5;6)<7:-q.")));
+        REQUIRE("#void:#range(#Range0,2,3),1<(#Range0+0),(#Range0+0)<4:b;"
+                "#void:4<5,5<7&#void:4<6,6<7:c:-q." == rewrite(parse("not 1<(2..3)<4:b|not 4<(5;6)<7:c:-q.")));
+        REQUIRE("b::1<2,2<3:-q." == rewrite(parse("b:1<2<3:-q.")));
+        REQUIRE("b::1>=2;"
+                "b::2>=3:-q." == rewrite(parse("b:not 1<2<3:-q.")));
+        // conjunctions
+        REQUIRE("h:-1<2&2<3:a." == rewrite(parse("h:-1<2<3:a.")));
+        REQUIRE("h:-a:1<2,2<3." == rewrite(parse("h:-a:1<2<3.")));
+        REQUIRE("h:-1>=2|2>=3:a." == rewrite(parse("h:-not 1<2<3:a.")));
+        REQUIRE("h:-a:2>=3;a:1>=2." == rewrite(parse("h:-a:not 1<2<3.")));
+        REQUIRE("h:-1<2&2<4|1<3&3<4:a." == rewrite(parse("h:-1<(2;3)<4:a.")));
+        REQUIRE("h:-a:1<3,3<4;a:1<2,2<4." == rewrite(parse("h:-a:1<(2;3)<4.")));
+        REQUIRE("h:-1>=2|2>=4|1>=3|3>=4:a." == rewrite(parse("h:-not 1<(2;3)<4:a.")));
+        REQUIRE("h:-a:3>=4;a:1>=3;a:2>=4;a:1>=2." == rewrite(parse("h:-a:not 1<(2;3)<4.")));
+        REQUIRE("h:-1<(#Range0+0)&(#Range0+0)<4&#range(#Range0,2,3):a." == rewrite(parse("h:-1<(2..3)<4:a.")));
+        REQUIRE("h:-a:1<(#Range0+0),(#Range0+0)<4,#range(#Range0,2,3)." == rewrite(parse("h:-a:1<(2..3)<4.")));
+        // Note: should be correct
+        REQUIRE("h:-1>=(#Range0+0)&#range(#Range0,2,3)|(#Range0+0)>=4&#range(#Range0,2,3):a." == rewrite(parse("h:-not 1<(2..3)<4:a.")));
+        REQUIRE("h:-"
+                "a:(#Range0+0)>=4,#range(#Range0,2,3),#range(#Range0,1,0);"
+                "a:1>=(#Range0+0),#range(#Range0,2,3),#range(#Range0,1,0)." == rewrite(parse("h:-a:not 1<(2..3)<4.")));
+        // set based head aggregates
+        REQUIRE("1<=#count{3:#void:1<2}:-q." == rewrite(parse("{ 1<2 } >= 1:-q.")));
+        REQUIRE("1<=#count{"
+                "3:#void:1<2;"
+                "4:#void:3<4"
+                "}:-q." == rewrite(parse("{ 1<2; 3<4 } >= 1:-q.")));
+        REQUIRE("1<=#count{"
+                "3:#void:1<2,2<4;"
+                "4:#void:1<3,3<4"
+                "}:-q." == rewrite(parse("{ 1<(2;3)<4 } >= 1:-q.")));
+        REQUIRE("#count{"
+                "3:#void:1>=2;"
+                "3:#void:2>=4;"
+                "4:#void:1>=3;"
+                "4:#void:3>=4"
+                "}:-q." == rewrite(parse("{ not 1<(2;3)<4 }:-q.")));
+        // Note the 3<=3 is the tuple and is meant to handle undefined terms
+        REQUIRE("#count{0:#void:}:-6<7;5<6;3<=3;q." == rewrite(parse("{ 5<6<7 }:-q.")));
+        // head aggreagets
+        REQUIRE("1<=#count{1:#void:p,1<2}:-q." == rewrite(parse("#count{ 1:1<2:p } >= 1:-q.")));
+        REQUIRE("1<=#count{1:#void:p,1<2,2<3}:-q." == rewrite(parse("#count{ 1:1<2<3:p } >= 1:-q.")));
+        REQUIRE("1<=#count{"
+                "1:#void:p,1>=2;"
+                "1:#void:p,2>=3"
+                "}:-q." == rewrite(parse("#count{ 1:not 1<2<3:p } >= 1:-q.")));
+        REQUIRE("1<=#count{"
+                "1:#void:p,1<2,2<4;"
+                "1:#void:p,1<3,3<4"
+                "}:-q." == rewrite(parse("#count{ 1:1<(2;3)<4:p } >= 1:-q.")));
+        REQUIRE("1<=#count{"
+                "1:#void:p,1>=2;"
+                "1:#void:p,2>=4;"
+                "1:#void:p,1>=3;"
+                "1:#void:p,3>=4"
+                "}:-q." == rewrite(parse("#count{ 1:not 1<(2;3)<4:p } >= 1:-q.")));
+        REQUIRE("1<=#count{"
+                "1:p:1<2,2<4;"
+                "1:p:1<3,3<4"
+                "}:-q." == rewrite(parse("#count{ 1:p:1<(2;3)<4 } >= 1:-q.")));
+        REQUIRE("1<=#count{"
+                "1:p:1>=2;"
+                "1:p:2>=4;"
+                "1:p:1>=3;"
+                "1:p:3>=4"
+                "}:-q." == rewrite(parse("#count{ 1:p:not 1<(2;3)<4 } >= 1:-q.")));
+        // set based body aggregates
+        REQUIRE("q:-1<=#count{3:1<2}." == rewrite(parse("q :- { 1<2 } >= 1.")));
+        REQUIRE("q:-1<=#count{"
+                "3:1<2;"
+                "4:3<4"
+                "}." == rewrite(parse("q :- { 1<2; 3<4 } >= 1.")));
+        REQUIRE("q:-1<=#count{"
+                "3:1<2,2<4;"
+                "4:1<3,3<4"
+                "}." == rewrite(parse("q :- { 1<(2;3)<4 } >= 1.")));
+        REQUIRE("q:-1<=#count{"
+                "3:1>=2;"
+                "3:2>=4;"
+                "4:1>=3;"
+                "4:3>=4"
+                "}." == rewrite(parse("q :- { not 1<(2;3)<4 } >= 1.")));
+        // body aggregates
+        REQUIRE("q:-1<=#count{1:1<2}." == rewrite(parse("q :- #count{ 1:1<2 } >= 1.")));
+        REQUIRE("q:-1<=#count{1:1<2,2<3}." == rewrite(parse("q :- #count{ 1:1<2<3 } >= 1.")));
+        REQUIRE("q:-1<=#count{"
+                "1:1>=2;"
+                "1:2>=3"
+                "}." == rewrite(parse("q :- #count{ 1:not 1<2<3 } >= 1.")));
+        REQUIRE("q:-1<=#count{"
+                "1:1<2,2<4;"
+                "1:1<3,3<4"
+                "}." == rewrite(parse("q :- #count{ 1:1<(2;3)<4 } >= 1.")));
+        REQUIRE("q:-1<=#count{"
+                "1:1>=2;"
+                "1:2>=4;"
+                "1:1>=3;"
+                "1:3>=4"
+                "}." == rewrite(parse("q :- #count{ 1:not 1<(2;3)<4 } >= 1.")));
+    }
+
+    SECTION("iesolve") {
+        // conjunction
+        REQUIRE("q:-p(X):1<=X,X<=3,#range(X,1,3)." ==  rewrite(parse("q :- p(X): 1 <= X <= 3.")));
+        // disjunction
+        REQUIRE("p(X)::1<=X,X<=3,#range(X,1,3)." ==  rewrite(parse("p(X): 1 <= X <= 3.")));
+        // head aggregate
+        REQUIRE("2<=#count{X:p(X):1<=X,X<=3,#range(X,1,3)}<=3." ==  rewrite(parse("2 #count { X: p(X): 1 <= X <= 3 } 3.")));
+        // body aggregate
+        REQUIRE("q:-2<=#count{0,p(X):1<=X,X<=3,p(X),#range(X,1,3)}<=3." ==  rewrite(parse("q :- 2 { p(X): 1 <= X <= 3 } 3.")));
+        REQUIRE("q:-2<=#count{X:p(X),1<=X,X<=3,#range(X,1,3)}<=3." ==  rewrite(parse("q :- 2 #count { X: p(X), 1 <= X <= 3 } 3.")));
+        // rule
+        REQUIRE("q:-#range(X,1,3);p(X);1<=X;X<=3." ==  rewrite(parse("q :- p(X), 1 <= X <= 3.")));
     }
 
     SECTION("defines") {
@@ -204,6 +353,10 @@ TEST_CASE("input-program", "[input]") {
         REQUIRE("#heuristic a[1@2,sign]:-[a]." == rewrite(parse("#heuristic a. [1@2,sign]")));
         REQUIRE("#heuristic a[1@2,level]:-[a]." == rewrite(parse("#heuristic a. [1@2,level]")));
         REQUIRE("#heuristic a[1@0,level]:-[a];c." == rewrite(parse("#heuristic a : c. [1,level]")));
+        REQUIRE("#external p:[#inc_base].[false]" == rewrite(parse("#external p.")));
+        REQUIRE("#external p(X):[#inc_base].[false]#external p(Y):[#inc_base].[false]" == rewrite(parse("#external p(X;Y).")));
+        REQUIRE("#external p:[#inc_base].[X]#external p:[#inc_base].[Y]" == rewrite(parse("#external p. [(X;Y)]")));
+        REQUIRE("#external p(X):[#inc_base].[X]#external p(X):[#inc_base].[Y]#external p(Y):[#inc_base].[X]#external p(Y):[#inc_base].[Y]" == rewrite(parse("#external p(X;Y). [(X;Y)]")));
     }
 
     SECTION("theory") {
